/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
//
//  XO2-256    modified ROM/RAM    XO2-256
//  27     B   1            24  VCC  
//  28     G   2         x  23  DIG1    25
//  29         3         x  22  DIG2    23
//  30         4            21  DIG3    21
//  32         5            20  DIG4    20
//   1         6          x 19  CLK_1HZ 17
//   4     C   7            18  BTN_DN  16
//   5     D   8          x 17  N.C.
//   8     E   9            16  BTN_UP  14
//   9     A   10           15  BTN_SEL 13
//  10     F   11           14  CLK     12
//      GND  12             13  CLK_OUT 11

//
//      A
//     ===
// F || G || B   DIG1 DIG2 DIG3 DIG4
//     ===
// E ||   || C
//     ===
//      D
//
// * anode common 
//    A-G    : Active Low
//    DIG1-4 : Active Low (need Inverter)
//                 __
//  BTN_SEL ------o  o---(R)--> VCC   (R : 0 or 1K )
//                 __
//  BTN_UP  ------o  o---(R)--> VCC   (R : 0 or 1K )
//                 __
//  BTN_DN  ------o  o---(R)--> VCC   (R : 0 or 1K ) , optical
//
`define SUPPORT_ALARM
`define SUPPORT_DN
//`define DIG_ACTIVE_HIGH
//`define SEG7_ACTIVE_HIGH
module clock # (
		//parameter CLK_HZ = 32768
		parameter CLK_HZ = 8000000
		//parameter CLK_HZ = 16000000
	) (
		input BTN_SEL
		,input BTN_UP
`ifdef SUPPORT_DN
		,input BTN_DN
`endif
                ,input CLK 
		,output CLK_OUT
                ,output A
                ,output B
                ,output C
                ,output D
                ,output E
                ,output F
                ,output G
                ,output CLK_1HZ

                ,output DIG1
                ,output DIG2
                ,output DIG3
                ,output DIG4
		);

    reg [15:0] pre_scaler = 0; // MAX 16 MHz
    reg [7:0] sub_sec = 0;

    wire clk_1hz_i = sub_sec[7];
    wire clk_8hz_i = sub_sec[4];
    wire CLK_128HZ = sub_sec[0];

    wire BLINK = clk_1hz_i;
    wire [1:0] DISP_SEL = sub_sec[2:1];
    wire [3:0] OUT;
    wire [3:0] sec_out;
    wire [3:0] min_out;
    wire [3:0] hour_out;
    reg [1:0] set_mode = 0; // 1: HH 2: MM 3: SS
`ifdef SUPPORT_ALARM
    reg alarm_mode = 0;
`else
    wire alarm_mode = 0;
`endif

    wire CLK_OUT = ~CLK;

    always @(negedge CLK)
    begin
	if (pre_scaler >= CLK_HZ/256 - 1)
	    begin
		pre_scaler <= 0 ;
		sub_sec <= sub_sec + 1 ;
	    end
	else    pre_scaler <= pre_scaler + 1;
    end

//                  BLINK==0    BLINK==1
//  set_mode == 0   HH:MM       HH:MM
//  set_mode == 1   HH:MM         :MM
//  set_mode == 2   HH:MM       HH:
//  set_mode == 3     :SS         :
//
    assign OUT [3:0] = BLINK &
			  ( (set_mode == 3) & DISP_SEL[1] 
			  | (set_mode == 1) & ~DISP_SEL[1]
			  | (set_mode == 2) & ~DISP_SEL[1] 
			  | (set_mode == 2) &  DISP_SEL[1] ) 
		      | ((set_mode == 3) & ~DISP_SEL[1]) ? 4'hF
			 : ( (set_mode == 3)  &  DISP_SEL[1]) ? sec_out[3:0]
			 : ( (set_mode == 1)  & ~DISP_SEL[1]) ? hour_out[3:0]
		   	 : min_out[3:0]
			 ;

	seg7dec 
`ifdef SEG7_ACTIVE_HIGH
		# ( .ACTIVE_HIGH(1) )
`endif
		seg7_impl (.A(A), .B(B), .C(C), .D(D), .E(E), .F(F), .G(G)
			, .DATA(OUT[3:0]) );
`ifdef DIG_ACTIVE_HIGH
    assign DIG1 = (DISP_SEL[1:0] == 0);
    assign DIG2 = (DISP_SEL[1:0] == 1);
    assign DIG3 = (DISP_SEL[1:0] == 2);
    assign DIG4 = (DISP_SEL[1:0] == 3);
`else
    assign DIG1 = ~(DISP_SEL[1:0] == 0);
    assign DIG2 = ~(DISP_SEL[1:0] == 1);
    assign DIG3 = ~(DISP_SEL[1:0] == 2);
    assign DIG4 = ~(DISP_SEL[1:0] == 3);
 `endif
 
    reg sec_up_i = 1'b0;
    wire min_up_i;
    wire hour_up_i;
	
    wire btn_sel_req;
    wire btn_up_req;
    reg btn_sel_ack = 1'b0;
    reg btn_up_ack = 1'b0;

	button btn_sel_impl (.CLK_128HZ(CLK_128HZ)
			, .I(BTN_SEL), .O(btn_sel_req), .R(btn_sel_ack) );
	button btn_up_impl  (.CLK_128HZ(CLK_128HZ)
			, .I(BTN_UP), .O(btn_up_req), .R(btn_up_ack) );

`ifdef SUPPORT_DN
    wire btn_dn_req;
    reg btn_dn_ack = 1'b0;

	button btn_dn_impl  (.CLK_128HZ(CLK_128HZ)
			, .I(BTN_DN), .O(btn_dn_req), .R(btn_dn_ack) );

    always @(posedge CLK_128HZ)
    begin
	btn_dn_ack <= btn_dn_req;
    end
`endif

    always @(posedge CLK_128HZ)
    begin
	sec_up_i <= (sub_sec == 8'hFF);
	btn_sel_ack <= btn_sel_req;
	btn_up_ack <= btn_up_req;
`ifdef SUPPORT_ALARM
	if (btn_sel_req)
	begin
 		set_mode <= (set_mode == 0) ? 1
 			   :(set_mode == 1) ? 2
 			   :(alarm_mode & (set_mode == 2)) ? 3
 			   :(set_mode == 3) ? 1
			   : 0
			   ;
 		alarm_mode <= (set_mode == 0) ? 1'b1
 			   :(alarm_mode & (set_mode == 1)) ? 1'b1
			   : 1'b0
			   ;
	end
`else
	if (btn_sel_req)
	begin
 		set_mode <= (set_mode == 0) ? 3
 			   :(set_mode == 3) ? 1
 			   :(set_mode == 1) ? 2
			   : 0
			   ;
	end
`endif
    end

`ifdef SUPPORT_DN
    //wire sec_dn =  (set_mode == 3) & btn_dn_ack ;
    wire sec_dn =  1'b0;
    wire min_dn =  (set_mode == 2) & btn_dn_ack ;
    wire hour_dn = (set_mode == 1) & btn_dn_ack ;
`else
    wire sec_dn =  1'b0;
    wire min_dn =  1'b0;
    wire hour_dn = 1'b0;
`endif
    //wire sec_up =  (set_mode == 3) & btn_up_ack ;
    wire sec_up =  1'b0;
    wire min_up =  (set_mode == 2) & btn_up_ack ;
    wire hour_up = (set_mode == 1) & btn_up_ack ;

	time_counter # ( .MAX(60) )
		sec_impl (.CLK(CLK_128HZ), .I_UP(sec_up_i), .O_UP(min_up_i)
		        , .I_UP_ALT(sec_up) , .I_DN(sec_dn)
			, .SEL({1'b0, DISP_SEL[0]}) , .O(sec_out[4:0]) );
	time_counter # ( .MAX(60) )
		min_impl (.CLK(CLK_128HZ), .I_UP(min_up_i), .O_UP(hour_up_i)
		        , .I_UP_ALT(min_up) , .I_DN(hour_dn)
			, .SEL({alarm_mode, DISP_SEL[0]}) , .O(min_out[3:0])
			, .MATCH(match_min) );
	time_counter # ( .MAX(24) )
		 hour_impl (.CLK(CLK_128HZ), .I_UP(hour_up_i)
		        , .I_UP_ALT(hour_up) , .I_DN(hour_dn) 
			, .SEL({alarm_mode, DISP_SEL[0]}), .O(hour_out[3:0])
			, .MATCH(match_hour) );

	assign CLK_1HZ = (match_hour & match_min) ? clk_8hz_i & clk_1hz_i
			: clk_1hz_i;
endmodule
